/*****************************************************************************
The main SFV code for use in SFV UNIX port

Copyright (C)1998 Alcatraz of Fantasia Sofware

The source code may not be modified or used in any other purpose then 
the compilation of sfv. It is distributed with the sfv package so it can be 
compiled on the desired platform. Any error or comments should be mailed to 

alc@cyberdude.com 
*****************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <dirent.h>
#include <sys/stat.h>
#include <ctype.h>
#include <signal.h>
#ifdef INCLUDEGETOPT
#include <getopt.h>
#endif
#include <fcntl.h>
#include <unistd.h>
#include "stack.h"
#include "sfvmisc.h"

extern int sfvaborted;

/*****************************************************************************
*pnShowHelp
*****************************************************************************/
void pnShowHelp()
{
  fprintf(stderr,
  	"\nUsage: sfv -[T|C|f filename|p filepath] options [filename(s)]\n"
	"  Operation mode controllers:\n"  
  	"      -T             To verify files against checksum table\n"
  	"      -C             To Create a checksum table\n"
	"  Common to both operation modes:\n"
  	"      -s             Does nothing in unix version, redirect to null instead\n"
  	"      -f filename    Forces SFV to use a specified filename rather then \n"
  	"                     autodetecting the name to use\n"
	"      -0             Turns off autodetection off filenames, causes sfv to use\n"
	"                     the file crc.txt\n"
  	"      -1             Use the first designated alternate checksum filename \n"
  	"                     *.txt (so releasename.txt is possible)\n"
  	"      -2             Same as -1 but with .sfv as extension. Default\n"
  	"      -3             Same as -1 but with .crc as extension\n"
	"      -p path        Path to the files if they are not in the cwd\n"
	"  Test mode only:\n"	
  	"      -l             Enable logging of bad files\n"
  	"      -n filename    Forces SFV to use a different log filename rather than\n"
  	"                     the default sfv.log for logging bad files\n"
  	"      -a             Append logfile rather then overwriting the existing\n"
	"      -d             Kill failed files. \n"
  	"  Create mode only:\n"
	"      -c filename    Add Filename as a comment to the checksum file.\n"
	"      -i             Add extened fileinfo in the checksum file\n"
	"      -o creator     Sets the given creator of the sfv file. Only in create mode\n"
  	"      -w             Inihibit showing of filecomment\n");
}

 


int main(int argc,char *argv[])
{
  	char 			*caszFileTypes[]={"Unknown type",
  					 "SFV32",
  					 "SFV32w",
  					 "SFV32nix",
  					 "early SFV32",
  					 "early SFV32w",
  					 "FileCRC-Linux port",
  					 "FileCRCw",
  					 "Validate"};

  	char			*caszIDStrings[]={"; Generated by SFV32 v1",
 	 				 "; Generated by WIN-SFV32 v",
  					 "; Generated by SFV32nix v1",
					 "; Generated by SFV32 v0",      //early
  					 "; Generated by WIN-SFV32 v0",  //early				 
   				     "Name Length      Date      Time      CRC-32      linux port",
   	                 "Name Length      Date      Time      CRC-32",
  					 "Directory of "};
  	char			*caszExtensions[]={".txt",".sfv",".crc"};
  
    	//Execution time calculators
  	time_t			sStartTime;
  	time_t			sEndTime;

   	 	//Program execution/argument handling
  	STACKRECORD		*MaskStack;
  	STACKRECORD		*LogStack;
  	STACKRECORD		*FileInfoStack;

  	char			*szCurrentWildCard=NULL;//current wildcard during processing
  	char			*szCurrentFile=NULL;	//current found filename
  	char			*szCurrentRow=NULL;  	//current line of file during reading
  	char 			szCurrentTimeString[30];  
  	DIR				*dp;
  	struct dirent	*dirp;
  	struct stat		statbuf;
  	char			fFileOK=1;
  	int  			iCurrentCRC;
  	int   			iCorrectCRC=0;
  	unsigned char	fileCRCFileNameWidth=13;
  	int				iErrors=0;		//number of errors found during processing
  	int				iIgnErrors=0;	//Number of ignored errors
	int				iSkipped=0;		//number of files in crc file but not processed
  	int				iExitCode=0;
  	int				isincomment=0;	//flag to tell if where displaying a comment or not .. 
  	long			iTotalFiles=0;
  	long			iTotalBytes=0;
  	preferences 	options;
  
    	//Files
  	FILE 			*extralogfile;
  	FILE			*listFile;
  	int				listFileType=0;
  	char			skipLines=0;

    	//General variables
  	int				i,j;
  	int				c;
  	char			*tmpStr=NULL;
  	char			*tmpStr2=NULL;  		//Always deallocate when done
  	STACKRECORD		*tmpStack;

  	struct tm		*sDisplayTime;
  	struct sigaction act;
  
		//Get program execution start time
  	time(&sStartTime);
  
  	if((sDisplayTime=gmtime(&sStartTime)))
  		strftime(szCurrentTimeString,30,"%Y-%m-%d at %H:%M.%S",sDisplayTime);
	else
		*szCurrentTimeString=0;

		//Allocate and initilalize variables
 	memset(&options,0,sizeof(preferences));
	MaskStack=stackCreate();
  	LogStack=stackCreate();
  	FileInfoStack=stackCreate();
  	*(tmpStr=malloc(PATH_MAX))=0;
  	options.logfilename=malloc(8);
  	strcpy(options.logfilename,"sfv.log");

		//display initial information
  	printf("%s - SimpleFileValidate Copyright (C)1998 Fantasia Software",PROGRAMVERSION);
  	printf("\nWritten by Alcatraz of Fantasia Software\n");

		//read the configuration file
  	if(loadpreferences(argv[0],&options))
		return EXITCODE_HELP;
  	
	SET_AUTOGETTYPE(options,2);
  	SET_TESTFILES(options);
  
    	//parse commandline
  	while((c=getopt(argc,argv,"CTihsldf:n:c:0123wop:ar"))!=EOF)
  	{
    	switch(c)
    	{
       		case 'C':  //Create file
	     		SET_CREATELIST(options);
         		SET_OKTOGO(options);
       		break;
       		case 'T':  //Test archives
	     		SET_TESTFILES(options);
		 		SET_OKTOGO(options);
       		break;
       		case 'i':  //Enable adding of fileinfo to log file
	     		SET_ADDFILEINFO(options);
       		break;
       		case 's':  //silent operation on
	     		SET_SILENT(options);
       		break;
       		case 'l':  //enable loggin
	     		SET_ENABLELOG(options);
       		break;
       		case 'd':  //delete failed files
	     		SET_KILLBAD(options);
       		break;
       		case 'f':  //set list file
	     		getstring(&options.crcfilename,optarg);
				SET_CRCNAMEGIVEN(options);
		 		SET_OKTOGO(options);
       		break;
       		case 'n':  //set log file
         		getstring(&options.logfilename,optarg); 
       		break;
       		case 'c':  //set comment file
	     		getstring(&options.commentfile,optarg);
       		break;
	   		case '0':  //dont autoget filename
       		case '1':  //txt
       		case '2':  //sfv
       		case '3':  //crc
	     		SET_AUTOGETTYPE(options,c-'0');
       		break;
       		case 'w':  //Dont show comment
	     		SET_NOCOMMENT(options);
       		break;
	   		case 'o':  //creator
	     		getstring(&options.creator,optarg);
	   		break;
	   		case 'p':  //filepath
	     		getpathname(&options.sitedir,optarg);
		 		SET_OKTOGO(options);
	   		break;
	   		case 'a':  //append logfile
	     		SET_APPENDLOGFILE(options);
	   		break;
	   		case 'r':  //recomment with default
	     		SET_RECOMMENTWITHDEFAULT(options);
	   		break;
       		case '?':
         		pnShowHelp();
         		return EXITCODE_HELP;
       		break;
   		}
  }
  		//no master parameter specified, abort
  	if(!OPTION_OKTOGO(options))
  	{
    	pnShowHelp();
    	return EXITCODE_HELP;
  	}
 
  		//dont autoget name, and no filename specified get the default name then
  	if(!OPTION_AUTOGETTYPE(options)&&!options.crcfilename)
  	{
     	options.crcfilename=realloc(options.crcfilename,(options.sitedir!=NULL?strlen(options.sitedir)+8:8));
	 	if(options.sitedir)
			strcpy(options.crcfilename,options.sitedir);
	 	else
			*options.crcfilename=0;
     	strcat(options.crcfilename,"crc.txt");
  	}

		/*	Dont use special operating mode when we have a filename and 
		 *	files specified */ 
	if(optind<argc)
		UNSET_CRCNAMEGIVEN(options);
  		//read all file arguments from commandline
  	while(optind<argc)
  	{
		if(!MaskStack&&!options.sitedir&&strrchr(argv[optind],'/'))
		{
			options.sitedir=realloc(options.sitedir,strlen(argv[optind])+1);
			strcpy(options.sitedir,argv[optind]);
			*(strrchr(options.sitedir,'/')+1)=0;
		}
    	stackPush(&MaskStack,argv[optind]);
    	if(OPTION_ADDFILEINFO(options)&&OPTION_CREATELIST(options)&&!stat(argv[optind],&statbuf))
    	{
      		strftime(tmpStr,50,"%H:%M.%S %Y-%m-%d",gmtime(&statbuf.st_mtime));
      		sprintf(&tmpStr[100],"; %13li %s %s",statbuf.st_size,tmpStr,argv[optind]);
      		stackPush(&FileInfoStack,tmpStr);
    	}
    	optind++;
  	}

   		//if no mask is specified default to all files
  	if(stackIsEmpty(MaskStack))
  	{  
		SET_NOMASKGIVEN(options);
		if(options.crcfilename&&!options.sitedir)
		{
			options.sitedir=realloc(options.sitedir,strlen(options.crcfilename)+1);
			strcpy(options.sitedir,options.crcfilename);
			if(strrchr(options.sitedir,'/'))
				*(strrchr(options.sitedir,'/')+1)=0;
			else{
				free(options.sitedir);
				options.sitedir=NULL;
			}
		}
		
    	strcpy(tmpStr,(options.sitedir?options.sitedir:"./"));
	  	i=(options.sitedir?strlen(options.sitedir):0);
		
		if((dp=opendir(tmpStr))!=NULL)
    	{
      		while((dirp=readdir(dp))!=NULL)
      		{
	    		strcpy(&tmpStr[i],dirp->d_name);
        		if(
           			strcmp(dirp->d_name,".")&&strcmp(dirp->d_name,"..")
           			&&
           			!stat(tmpStr,&statbuf)&&!S_ISDIR(statbuf.st_mode)
          		  )
        		{
          			stackPush(&MaskStack, tmpStr);
          			if(OPTION_ADDFILEINFO(options)&&OPTION_CREATELIST(options)&&!stat(tmpStr,&statbuf))
          			{                                              
            			strftime(tmpStr,50,"%H:%M.%S %Y-%m-%d",gmtime(&statbuf.st_mtime));
            			sprintf(&tmpStr[100],"; %13li %s %s",statbuf.st_size,tmpStr,dirp->d_name);
            			stackPush(&FileInfoStack,tmpStr);
          			}
        		}
      		}
      		closedir(dp);
    	}else
      	{
			tmpStr2=malloc(strlen(tmpStr)+5);
	   		sprintf(tmpStr2,"%s: ",argv[0]);
	    	perror(tmpStr2);
			free(tmpStr2);
        	return EXITCODE_IOERR;
      	}
  	}
    	//attempt to autogetfilename based upon files found
  	if(OPTION_AUTOGETTYPE(options)&&!options.crcfilename&&!stackIsEmpty(MaskStack))
  	{
    	tmpStack=MaskStack;
    	while(tmpStack!=NULL)
    	{
      		strcpy(tmpStr,tmpStack->data);
      		if((i=strlen(tmpStr))>4)
      		{
	    		for(i=strlen(tmpStr)-3;i<strlen(tmpStr);i++)
					tmpStr[i]=toupper(tmpStr[i]);
        		if(
					!strcmp(&tmpStr[i-4],".RAR")
            		||
            		(
			  			(tmpStr[i-4]=='.')
			  			&&
              			(tmpStr[i-3]=='R'||tmpStr[i-3]=='S'||(tmpStr[i-3]>='0'&&tmpStr[i-3]<='9'))
              			&&
              			(tmpStr[i-2]>='0'&&tmpStr[i-2]<='9')
              			&&
              			(tmpStr[i-1]>='0'&&tmpStr[i-1]<='9')
            		)
          		  )
        		{
          			strcpy(&tmpStr[i-4],caszExtensions[OPTION_AUTOGETTYPE(options)-1]);
          			options.crcfilename=realloc(options.crcfilename,i+1);
          			strcpy(options.crcfilename,tmpStr);
          			break;
        		}
      		}
      		tmpStack=tmpStack->next;
    	}
  	}
  
	  	//if we didnt get the filename by now, try scanning for all files matching our extensions
  	if(
		OPTION_AUTOGETTYPE(options)
		&&
		(
		 	!options.crcfilename
		 	||
		 	(OPTION_TESTFILES(options)&&!ffnFileExists(options.crcfilename))
		)
		&&
	   	!scanforcrcfile(&options.crcfilename,caszExtensions[1],(options.sitedir?options.sitedir:"./"),argv[0])
		&&
	   	!scanforcrcfile(&options.crcfilename,caszExtensions[2],(options.sitedir?options.sitedir:"./"),argv[0])
		&&
	   	!scanforcrcfile(&options.crcfilename,caszExtensions[0],(options.sitedir?options.sitedir:"./"),argv[0])
	  )
  	{
    	fprintf(stderr,"%s: Unable to autoget the listfilename, using default\n",argv[0]);
    	options.crcfilename=realloc(options.crcfilename,(options.sitedir!=NULL?strlen(options.sitedir)+8:8));
		if(options.sitedir)
			strcpy(options.crcfilename,options.sitedir);
   		else
   			*options.crcfilename=0;
  		strcat(options.crcfilename,"crc.txt");
  	}

  	printf("- Using %s as sfv file\n",options.crcfilename);
 
  		//set up the break handler during processing
  	act.sa_handler=handleSIGINT;
  	act.sa_flags=SA_RESTART;
  	sigemptyset(&act.sa_mask);
  	sigaction(SIGINT,&act,NULL);
  
  if(OPTION_TESTFILES(options))
  {
    szCurrentRow=malloc(MAXROWLENGTH);
	
      //Check files against table
    if((listFile=fopen(options.crcfilename,"r"))!=NULL)
    {
      szCurrentWildCard=malloc(MAXROWLENGTH);
      while(fgets(szCurrentRow,MAXROWLENGTH-1,listFile)!=NULL)
      {
        fFileOK=0;
          //note this also trims the actual string
        if(!skipLines&&strlen(szfnTrim(szCurrentRow)))
        {
          switch(listFileType)
          {
            case 0:  //first off we shall determine file type
              for(i=0;i<8;i++)
              {
                if(i<5||i>6)
                {
                  if(strncmp(szCurrentRow,caszIDStrings[i],strlen(caszIDStrings[i]))==0)
                  {
                      //listFileType=0 == iunknown so theese indexes are +1
                    listFileType=i+1;
                    skipLines=2;
                    break;
                  }
                }else
                  {
                    if(strncmp(szCurrentRow,caszIDStrings[i],5)==0)
                    {
                      for(j=4;j<24;j++)
                      {
                        if(szCurrentRow[j]=='L'&&strncmp(&szCurrentRow[j],&caszIDStrings[i][5],strlen(caszIDStrings[i])-6)==0)
                        {
                          listFileType=i+1;
                          skipLines=2;
                          break;
                        }else
                          {
                            if(szCurrentRow[j]!=' '||szCurrentRow[j]==0)
                              break;
                          }
                      }
                      if(listFileType)
                        break;
                    }
                  }
              }
              if(listFileType)
              {
                printf("- Identified input file as %s\n",caszFileTypes[listFileType]);
                if(listFileType==4||listFileType==5)
                {
                  fprintf(stderr,"%s: The unix port cannot read theese early checksums, please optain a checksum file created by a later version of sfv or download any of the Win32 versions\n",argv[0]);
                  return EXITCODE_HELP;
                }
              }
            break;
            case 1:  //SFV32
            case 2:  //SFV32w
            case 3:  //SFV32nix
              if(szCurrentRow[0]!=';')
              {
				if(strlen(szCurrentRow)>9)
				{
					strncpy(tmpStr,&szCurrentRow[strlen(szCurrentRow)-8],8);
					tmpStr[8]=0;
                	iCorrectCRC=strtoul(tmpStr,(char **)NULL,16);
                	strncpy(szCurrentWildCard,szCurrentRow,strlen(szCurrentRow)-9);
                	szCurrentWildCard[strlen(szCurrentRow)-9]=0;
                	fFileOK=1;
				}
              }else
			  {
				if(isincomment)
				{
					if(strcmp(szCurrentRow,COMMENTEND))
						printf("%s\n",&szCurrentRow[1]);
					else
						isincomment=0;
				}
				else
				{
					if(!strcmp(szCurrentRow,COMMENTSTART)&&!OPTION_NOCOMMENT(options))
						isincomment=1;
					else
						if(!strncmp(szCurrentRow,CREATORSTART,strlen(CREATORSTART))&&strlen(szCurrentRow)>strlen(CREATORSTART)+1)
							printf("- sfv file created by %s\n",strcpy(tmpStr,&szCurrentRow[strlen(CREATORSTART)+1]));
				}
			  }
            break;
            case 6:  //FileCrc linux
            case 7:  //FileCRCw
			  if(strlen(szCurrentRow)>9)
			  {
              	iCorrectCRC=strtoul(strncpy(tmpStr,&szCurrentRow[strlen(szCurrentRow)-8],8),(char **)NULL,16);
              	strncpy(szCurrentWildCard,szCurrentRow,fileCRCFileNameWidth);
              	szCurrentWildCard[fileCRCFileNameWidth]=0;
              	szfnTrim(szCurrentWildCard);
              	fFileOK=1;
			  }
            break;
            case 8:  //Validate
              if(strlen(szCurrentRow)>40)
              {
                strncpy(tmpStr,&szCurrentRow[43],4);
				strncpy(&tmpStr[4],&szCurrentRow[50],4);
				tmpStr[8]=0;
			    iCorrectCRC=strtoul(tmpStr,(char **)NULL,16);
				
				strncpy(szCurrentWildCard,szCurrentRow,8);
                szCurrentWildCard[8]=0;
                szfnTrim(szCurrentWildCard);
                strncpy(&szCurrentWildCard[(i=strlen(szCurrentWildCard)+1)],&szCurrentRow[9],3);
                szCurrentWildCard[i-1]='.';
                szCurrentWildCard[i+3]=0;
                szfnTrim(szCurrentWildCard);
                fFileOK=1;
              }
            break;
          }
            //fFileOK tells me wether or not we have a valid file this round
          if(fFileOK) 
          {  
             //is the file in the file but not specifed for test?
            if((szCurrentFile=szfnScanStackForFile(MaskStack,szCurrentWildCard)))
            {
              iCurrentCRC=ifnGetFileCRC(szCurrentFile,&fFileOK,listFileType,&iTotalFiles,&iTotalBytes,argv[0]);
              tmpStr2=szfnGetFilePart(szCurrentFile);
			  if((OPTION_NOMASKGIVEN(options)||OPTION_CRCNAMEGIVEN(options)?iTotalFiles+iSkipped:iTotalFiles)==1)
				printf("\n");
			  if(fFileOK)
              {
                if(iCurrentCRC==iCorrectCRC)
                {
                  printf(" [   OK   ] - %s\n",tmpStr2);
                }else
                  {
                   	printf(" [  BAD   ]");
					if(!isignoreext(szCurrentFile,&options))
					{
						iExitCode=EXITCODE_FAILED;
                   		iErrors++;
                    	if(OPTION_KILLBAD(options))
                    	{
                      		if(remove(szCurrentFile))
                      		{
                       	 		printf(" - %s\n",tmpStr2);
                        		sprintf(tmpStr,"%s: %s\n",argv[0],szCurrentFile);
								perror(tmpStr);
                      		}else
                        		printf("[KILLED] - %s\n",tmpStr2);
                       
                    	}else
						{
							if(options.badfileappendix)
							{
								if(rename(szCurrentFile,strcat(strcpy(tmpStr,szCurrentFile),(options.badfileappendix?options.badfileappendix:".bad"))))
								{
							    	printf(" - %s\n",tmpStr2);
									sprintf(tmpStr,"%s: %s",argv[0],szCurrentFile);
									perror(tmpStr);
								}else
									printf("[RENAMED] - %s\n",tmpStr2);
							}else
								printf(" - %s\n",tmpStr2);
						}
                    
                    	sprintf(tmpStr,"%s failed tests. CRC was %08X but should be %08X",szCurrentFile,iCurrentCRC,iCorrectCRC);
                    	stackPush(&LogStack,tmpStr);
					}else
					{
						iIgnErrors++;
						printf("[IGNORED] - %s\n",tmpStr2);
					}
                  }
              }else
                {
					if(!sfvaborted)
					{
                  		sprintf(tmpStr,"%s could not be opened",szCurrentFile);
                  		stackPush(&LogStack,tmpStr);
                  		iExitCode=EXITCODE_FAILED;
				  		printf(" [IOERROR ] - %s\n",tmpStr2);
                  		iErrors++;
					}
                }
            }else
			{
				if(
					OPTION_CRCNAMEGIVEN(options)
					||
					OPTION_NOMASKGIVEN(options)
				  )
				{
              		if(!isignoreext(szCurrentWildCard,&options))
					{
						iSkipped++;
						if(OPTION_CREATEMISSINGFILE(options))
						{
							tmpStr[0]=0;
							if(options.sitedir)
								strcpy(tmpStr,options.sitedir);
							strcat(tmpStr,szCurrentWildCard);
							if((i=open(strcat(tmpStr,(options.missingappendix?options.missingappendix:".missing")),O_WRONLY|O_CREAT|O_EXCL,0))==-1)
							{
								fprintf(stderr,"%s:",argv[0]);
								perror(tmpStr);
							}else
								close(i);
						}
					}else
						iIgnErrors++;
					if(iSkipped+iTotalFiles+iIgnErrors==1)
						printf("\n");
					printf(" [MISSING ] - %s\n",szCurrentWildCard);
				}
			}
          }
        }
        if(skipLines)
        {
          skipLines--;
          if(!skipLines&&(listFileType==6||listFileType==7)&&strlen(szCurrentRow))
          {
            for(fileCRCFileNameWidth=0;fileCRCFileNameWidth<strlen(szCurrentRow)&&szCurrentRow[fileCRCFileNameWidth]=='-';fileCRCFileNameWidth++);
          }
        }
		if(sfvaborted)
		{
			printf("Aborted.\n");
			break;
		}
      }
      if(!listFileType)
      {
        fprintf(stderr,"%s: Unable to identify the type of the sfv file\n",argv[0]);
        iExitCode=EXITCODE_IOERR;
      }else
	  {
      	if(iErrors) printf("%i Error(s)",iErrors);
      	else
        	printf("All OK");
      	if(iSkipped) printf(". %d File(s) skipped",iSkipped);
		if(iIgnErrors) printf(". %d Error(s) ignored",iIgnErrors);
     	printf("\n");

		
	  	if(
			(!iSkipped)
			&&
			(
				(iErrors&&options.createtestbad)
				||
				(
			 		!iErrors
					&&
					options.createtestok
					&&
					(
					 	OPTION_NOMASKGIVEN(options)
						||
						!iSkipped
					)
				)
			)
		  )
	  	{
			if(options.sitedir)
				strcpy(tmpStr,options.sitedir);
			else
				*tmpStr=0;
			if(iErrors)
				strcat(tmpStr,options.createtestbad);
			else
				strcat(tmpStr,options.createtestok);
			if(mkdir(tmpStr,0755))
			{
				tmpStr2=malloc(strlen(argv[0])+strlen(tmpStr)+10);
				sprintf(tmpStr2,"%s: %s",argv[0],tmpStr);
				perror(tmpStr2);
				free(tmpStr2);
			}
	  	}
	  }
	  
      free(szCurrentWildCard);
      fclose(listFile);
	  
        //output data to a logfile, if loggin is enabled and we have somthing to write
      if(
			(
			 	(
			  		OPTION_ENABLELOG(options)
			  		&&
					options.logfilename
				)
			 	||
				options.badtestlogfilename
			)
			&&
			!stackIsEmpty(LogStack))
      {
 		if(OPTION_ENABLELOG(options)&&options.logfilename)
		{
			if(!(listFile=fopen(options.logfilename,(OPTION_APPENDLOGFILE(options)?"a":"w"))))
			{
				sprintf(tmpStr,"%s: %s",argv[0],options.logfilename);
				perror(tmpStr);
			}
		}else
			listFile=NULL;
		if(options.badtestlogfilename)
		{
			if(options.badtestlogfilename[0]!='/'&&options.sitedir)
			{
				strcpy(tmpStr,options.sitedir);
				strcat(tmpStr,options.badtestlogfilename);
			}else
				strcpy(tmpStr,options.badtestlogfilename);
			if(!(extralogfile=fopen(tmpStr,"w")))
			{
				tmpStr2=malloc(strlen(tmpStr)+strlen(argv[0])+10);
				sprintf(tmpStr2,"%s: %s",argv[0],tmpStr);
				perror(tmpStr2);
				free(tmpStr2);
			}
		}else
			extralogfile=NULL;
        if(extralogfile||listFile)
        {
		  if(listFile)
          	fprintf(listFile,"Generated by %s on %s \n",PROGRAMVERSION, szCurrentTimeString);
		  if(extralogfile)
          	fprintf(extralogfile,"Generated by %s on %s \n",PROGRAMVERSION, szCurrentTimeString);
			
          while(!stackIsEmpty(LogStack))
          {
			tmpStr2=stackPop(&LogStack);
			if(listFile)
            	fprintf(listFile,"%s\n",tmpStr2);
			if(extralogfile)
				fprintf(extralogfile,"%s\n",tmpStr2);
            free(tmpStr2);
          }
		  if(listFile)
          	fclose(listFile);
		  if(extralogfile)
			fclose(extralogfile);
        }
      }
      free(szCurrentRow);
    }else
      {
		sprintf(tmpStr,"%s: %s",argv[0],options.crcfilename);
        perror(tmpStr);
        return EXITCODE_FILEOPENERROR;
      }
    
  }else
    { 
      //Create checksum table
      if((listFile=fopen(options.crcfilename,"w"))!=NULL)
      {
          //first off we output the file header
        fprintf(listFile,"; Generated by %s on %s\n;\n",PROGRAMVERSION,szCurrentTimeString);
        if(options.creator)
			fprintf(listFile,"%s %s\n",CREATORSTART,options.creator);
		if(options.commentfile&&!ffnAddComment(listFile,options.commentfile))
          return EXITCODE_FILEOPENERROR;
        
		while(!stackIsEmpty(FileInfoStack))
        {
          fprintf(listFile,"%s\n",(tmpStr2=stackPop(&FileInfoStack)));
          free(tmpStr2);
        }
        printf("\n");
          //Loop til we run out of files
        while(!stackIsEmpty(MaskStack))
        {
          szCurrentWildCard=stackPop(&MaskStack);
          iCurrentCRC=ifnGetFileCRC(szCurrentWildCard,&fFileOK,3,&iTotalFiles,&iTotalBytes,argv[0]);
           
		  tmpStr2=szfnGetFilePart(szCurrentWildCard);
          if(fFileOK)
          {
            printf(" [%08X] - %s\n",iCurrentCRC,tmpStr2);   
            fprintf(listFile,"%s %08X\n",tmpStr2,iCurrentCRC);
          }else
            {
              printf(" [IOERROR ] - %s\n", tmpStr2);
              iExitCode=EXITCODE_IOERR;
            }
          free(szCurrentWildCard);
		  if(sfvaborted)
		  {
			printf("\nAborted.\n");
			return EXITCODE_CANCELED;
		  }
        }
        fclose(listFile);
		if(sfvaborted)
		{
			if(remove(options.crcfilename))
			{
				sprintf(tmpStr,"%s: %s",argv[0],options.crcfilename);
				perror(tmpStr);
			}
		}
      }else
        {
          sprintf(tmpStr,"%s: %s",argv[0],options.crcfilename);
		  perror(tmpStr);
          return EXITCODE_FILEOPENERROR;
        }
    }
    //free up most variables
  stackDispose(&MaskStack);
  stackDispose(&LogStack);

    //Calculate and display program execution time
  time(&sEndTime);
  sEndTime=sEndTime-sStartTime;
  sDisplayTime=gmtime(&sEndTime);
  strftime(tmpStr,255,"Active %H:%M.%S",sDisplayTime);
  printf("\n%s\n %li byte(s) in %li file(s), %.2f kb/s\n",tmpStr,iTotalBytes,iTotalFiles,(double)((iTotalBytes/(sEndTime?sEndTime:1))/1024));
  if(iTotalFiles==0) iExitCode=EXITCODE_NOTHING;

  return iExitCode;
}
