/*****************************************************************************
Misc functions for use in SFV UNIX port

Copyright (C)1998 Alcatraz of Fantasia Sofware

The source code may not be modified or used in any other purpose then 
the compilation of sfv. It is distributed with the sfv package so it can be 
compiled on the desired platform. Any error or comments should be mailed to 

alc@cyberdude.com
*****************************************************************************/
#include <limits.h>

#ifndef PATH_MAX
#define PATH_MAX 32768
#endif

#define MAXROWLENGTH 8192

	//whether or not to use fread/fopen rather then read/open while testing 
//#define USESTDIO

//Versionstring to display on misc occasions
#ifdef SFV
  #define PROGRAMVERSION "SFV32nix v1.2"
  #define USERSETTINGSFILE "/.sfvrc"
  #define GLOBALSETTINGSFILE "/etc/sfvrc"
#else
  #define PROGRAMVERSION "SFV32Testnix v1.1a"
  #define USERSETTINGSFILE "/.sfvtestrc"
  #define GLOBALSETTINGSFILE "/etc/sfvtestrc"
#endif
	//comment start and stop commands
#define COMMENTSTART ";!SFV_COMMENT_START"
#define COMMENTEND	";!SFV_COMMENT_END"
#define CREATORSTART ";!SFV_CREATOR"

  	//buffersize used when reading filedata for crc calculation
#define READBUFSIZE 32768*4

  	//Program exitcodes         
#define EXITCODE_FILEOPENERROR 0x01                        
#define EXITCODE_FAILED	0x02
#define EXITCODE_IOERR	0x03
#define EXITCODE_HELP	0x04
#define EXITCODE_NOTHING	0x05
#define EXITCODE_BADLOG	0x06
#define EXITCODE_CANCELED 0x07
 
#define OPTION_CREATETESTBADLOG(b) ((b).miscoptions1&0x01)
#define OPTION_TESTFILES(b) ((b).miscoptions1&0x02)
#define OPTION_CREATELIST(b) ((b).miscoptions1&~0x02)
#define OPTION_OKTOGO(b) ((b).miscoptions1&0x04)
#define OPTION_SILENT(b) ((b).miscoptions1&0x08)
#define OPTION_ENABLELOG(b) ((b).miscoptions1&0x10)
#define OPTION_KILLBAD(b) ((b).miscoptions1&0x20)
#define OPTION_NOCOMMENT(b) ((b).miscoptions1&0x40)
#define OPTION_ADDFILEINFO(b) ((b).miscoptions1&0x80)
#define OPTION_RECOMMENTLIST(b) ((b).miscoptions1&0x100)
#define OPTION_APPENDLOGFILE(b) ((b).miscoptions1&0x200)
#define OPTION_RECOMMENTWITHDEFAULT(b) ((b).miscoptions1&0x400)
#define OPTION_CRCNAMEGIVEN(b) ((b).miscoptions1&0x800)
#define OPTION_NOMASKGIVEN(b) ((b).miscoptions1&0x1000)
#define OPTION_CREATEMISSINGFILE(b) ((b).miscoptions1&0x2000)
#define OPTION_AUTOGETTYPE(b) ((b).autogettype)

#define SET_CREATETESTBADLOG(b) ((b).miscoptions1|=0x01)
#define SET_TESTFILES(b) ((b).miscoptions1|=0x02)
#define SET_CREATELIST(b) ((b).miscoptions1&=~0x02)
#define SET_OKTOGO(b) ((b).miscoptions1|=0x04)
#define SET_SILENT(b) ((b).miscoptions1|=0x08)
#define SET_ENABLELOG(b) ((b).miscoptions1|=0x10)
#define SET_KILLBAD(b) ((b).miscoptions1|=0x20)
#define SET_NOCOMMENT(b) ((b).miscoptions1|=0x40)
#define SET_ADDFILEINFO(b) ((b).miscoptions1|=0x80)
#define SET_RECOMMENTLIST(b) ((b).miscoptions1|=0x100)
#define SET_APPENDLOGFILE(b) ((b).miscoptions1|=0x200)
#define SET_RECOMMENTWITHDEFAULT(b) ((b).miscoptions1|=0x400)
#define SET_CRCNAMEGIVEN(b) ((b).miscoptions1|=0x800)
#define SET_NOMASKGIVEN(b) ((b).miscoptions1|=0x1000)
#define SET_CREATEMISSINGFILE(b) ((b).miscoptions1|=0x2000)
#define SET_AUTOGETTYPE(b,c) ((b).autogettype=c)

#define UNSET_APPENDLOGFILE(b) ((b).miscoptions1&=~0x200)
#define UNSET_CREATETESTBADLOG(b) ((b).miscoptions1&=~0x200)
#define UNSET_CREATEMISSINGFILE(b) ((b).miscoptions1&=~0x2000)
#define UNSET_CRCNAMEGIVEN(b) ((b).miscoptions1&=~0x800)

typedef struct {
		char	*sitedir;
		char	*crcfilename;
		char	*logfilename;
		char	*createtestok;
		char	*createtestbad;
		char	*badfileappendix;
		char	*badtestlogfilename;
		char	*commentfile;
		char	*creator;
		char	*ignoreext;
		char	*missingappendix;
		int	    autogettype;
		int		miscoptions1;
} preferences;


  	//returns calculated crc of file valid if fFileOK!=0
long ifnGetFileCRC(char *szFile,char *fFileOK,char fileType,long *iTotalFiles,long *iTotalBytes,char *argv0);
  	//calculates a crc32 of the given buffer
long crc32r(unsigned long iCRC,const char *buff,long count);
  	//trims input string pff leading and trailing spaces, returning a pointer to input
char *szfnTrim(char *szInput);
  	//scans the stack for a mathing file and returns a pointer to the actual data in the stack
char *szfnScanStackForFile(STACKRECORD *stack,char *szFile);
	//returns filepart of argument
char *szfnGetFilePart(char *szPath);
  	//scans the given folder for a file with the given extension
char *scanforcrcfile(char **szListFile,char *szExtension,char *szUseDir,char *szArgv0);
 	//loads prefernces from a file
int loadpreferences(char *programname,preferences *prefs);
  	//Verifies the readability/existance of a file
char ffnFileExists(char *szFile);
  	//Adds a commentfile to the already opened listfile
char ffnAddComment(FILE *listFile,char *szCommentFile);
  	//reallocs buf according to lentgh of path, also adds a slash to path in buf
void getpathname(char **buf,char *path);
  	//does same as getpathname, but no backslash
void getstring(char **buf,char *str);
  	//sets aborted to true upon ctrl-break
void handleSIGINT(int signo);
	//Verifies if the given filenames error can be ignored
int isignoreext(char *fname,preferences *prefs);
